from .default_model import DefaultLightGBMModel, DefaultXGBoostModel, DefaultCatBoostModel, DefaultGenericModel
from .sample_model import SampleModel
from .sample_rule_model import SampleRuleModel
from .info_unpaid_rule_model import InfoUnpaidRuleModel
from .info_defect_rule_model import InfoDefectRuleModel
from .vas_xcloud_rule_model import VasXcloudRuleModel
from .generic_context_model import GenericContextModel
from .generic_logic_model import GenericLogicModel
from .conditional_generic_logic_model import ConditionalGenericLogicModel
from .tw_random_greeting_rule_model import TwRandomGreetingRuleModel
from .lightgbm_device_model import LightGBMDeviceModel
from .sample_pytorch_model import SamplePyTorchModel
from .catboost_device_model import CatboostDeviceModel
from .commbot_catboost_device_model import CommbotCatboostDeviceModel
from .eureka_model import EurekaModel

__all__ = [
    "DefaultLightGBMModel",
    "DefaultXGBoostModel",
    "DefaultCatBoostModel",
    "DefaultGenericModel",
    "SampleModel",
    "SampleRuleModel",
    "SamplePyTorchModel",
    "InfoUnpaidRuleModel",
    "InfoDefectRuleModel",
    "VasXcloudRuleModel",
    "GenericContextModel",
    "GenericLogicModel",
    "ConditionalGenericLogicModel",
    "TwRandomGreetingRuleModel",
    "LightGBMDeviceModel",
    "CatboostDeviceModel",
    "CommbotCatboostDeviceModel",
    "EurekaModel",
]



import asyncio
import importlib.util
import re
from abc import ABCMeta, abstractmethod
from functools import partial
from pathlib import Path
from typing import Any, Callable, Dict, List, Union

import autogluon.tabular.version
import catboost
import lightgbm
import torch
import xgboost

from sktmls import ModelRegistry, MLSENV, MLSRuntimeENV


class MLSModelError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class MLSModel(metaclass=ABCMeta):
    """
    MLS 모델 레지스트리에 등록되는 최상위 클래스입니다.
    """

    def __init__(self, model_name: str, model_version: str):
        assert type(model_name) == str
        assert type(model_version) == str

        if not bool(re.search("^[A-Za-z0-9_]+_model$", model_name)):
            raise MLSModelError(
                "model_name should follow naming rule. MUST be in alphabet, number, underscore and endwiths '_model'"
            )

        if not bool(re.search("^[A-Za-z0-9_]+$", model_version)):
            raise MLSModelError("model_name should follow naming rule. MUST be in alphabet, number, underscore")

        self.models = None
        self.model_name = model_name
        self.model_version = model_version
        self.model_lib = None
        self.model_lib_version = None
        self.features = None

    def save(self, env: MLSENV = None, runtime_env: MLSRuntimeENV = None, force: bool = False) -> None:
        """
        모델 바이너리(model.joblib)와 정보(model.json)를 MLS 모델 레지스트리에 등록합니다.

        `sktmls.ModelRegistry.save`와 동일하게 동작합니다.

        ## Args

        - env: (`sktmls.MLSENV`) 접근할 MLS 환경 (`sktmls.MLSENV.DEV`|`sktmls.MLSENV.STG`|`sktmls.MLSENV.PRD`) (기본값: `sktmls.MLSENV.STG`)
        - runtime_env: (`sktmls.MLSRuntimeENV`) 클라이언트가 실행되는 환경 (`sktmls.MLSRuntimeENV.YE`|`sktmls.MLSRuntimeENV.EDD`|`sktmls.MLSRuntimeENV.LOCAL`) (기본값: `sktmls.MLSRuntimeENV.LOCAL`)
        - force: (bool) 이미 모델 레지스트리에 등록된 경우 덮어 쓸 것인지 여부 (기본값: `False`)
        """
        model_registry = ModelRegistry(env=env, runtime_env=runtime_env)
        model_registry.save(self, force=force)

    @abstractmethod
    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        pass

    async def predict_async(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        return await asyncio.get_event_loop().run_in_executor(None, partial(self.predict, x=x, **kwargs))


class MLSLightGBMModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 LightGBM 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str] = None):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "lightgbm"
        self.model_lib_version = lightgbm.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSXGBoostModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 XGBoost 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "xgboost"
        self.model_lib_version = xgboost.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSCatBoostModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 CatBoost 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "catboost"
        self.model_lib_version = catboost.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSRuleModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 Rule 기반 상위 클래스입니다.
    """

    def __init__(self, model_name: str, model_version: str, features: List[str]):
        super().__init__(model_name, model_version)
        self.model_lib = "rule"
        self.model_lib_version = "N/A"

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")


class MLSGenericModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 제네릭 모델 기반 상위 클래스입니다.
    """

    def __init__(
        self,
        models,
        model_name: str,
        model_version: str,
        features: List[str],
        custom_predict_async: Union[str, Callable] = None,
    ):
        assert type(models) == list, "`models`는 list 타입이어야 합니다."

        super().__init__(model_name, model_version)
        self.models = models
        self.model_lib = "etc"
        self.model_lib_version = "N/A"

        if not custom_predict_async:
            self.custom_predict_async = None
        elif isinstance(custom_predict_async, str):
            self.load_custom_predict_async(custom_predict_async)
        else:
            self.custom_predict_async = custom_predict_async

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")

    def set_model_lib(self, model_lib: str):
        """
        모델이 사용한 라이브러리 정보를 세팅하기 위한 함수입니다.
        """
        self.model_lib = model_lib
        if model_lib == "lightgbm":
            self.model_lib_version = lightgbm.__version__
        elif model_lib == "catboost":
            self.model_lib_version = catboost.__version__
        elif model_lib == "xgboost":
            self.model_lib_version = xgboost.__version__
        elif model_lib == "pytorch":
            self.model_lib_version = torch.__version__
        elif model_lib == "autogluon":
            self.model_lib_version == autogluon.tabular.version.__version__
        elif model_lib == "etc" or model_lib == "rule":
            self.model_lib_version == "N/A"

    def load_custom_predict_async(self, file_path: str = None) -> Callable:
        file_path = file_path or Path.home().joinpath("models").joinpath(
            self.model_name, self.model_version, "predict_async.py"
        )
        try:
            spec = importlib.util.spec_from_file_location(f"{self.model_name}/{self.model_version}", file_path)
            m = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(m)
            self.custom_predict_async = m.predict_async
        except (FileNotFoundError, AttributeError):
            self.custom_predict_async = None

    def predict(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        return {"items": []}

    async def predict_async(self, x: List[Any], **kwargs) -> Dict[str, Any]:
        if hasattr(self, "custom_predict_async") and self.custom_predict_async is not None:
            return await self.custom_predict_async(x, **kwargs)
        return await asyncio.get_event_loop().run_in_executor(None, partial(self.predict, x=x, **kwargs))


class MLSPyTorchModel(MLSModel):
    """
    MLS 모델 레지스트리에 등록되는 단일 PyTorch 기반 상위 클래스입니다.
    """

    def __init__(self, model, model_name: str, model_version: str, features: List[str]):
        assert isinstance(model, torch.jit.ScriptModule), "`model`은 `torch.jit.ScriptMoudle` 타입이어야 합니다"
        assert torch.__version__.split("+")[0] == "1.8.1", "`torch`버전은 1.8.1 이어야 합니다."

        super().__init__(model_name, model_version)
        self.models = [model]
        self.model_lib = "pytorch"
        self.model_lib_version = torch.__version__

        if features:
            self.features = features
        else:
            raise MLSModelError("`features`가 없습니다.")



from pathlib import Path
from typing import Dict, List

import pandas as pd
from autogluon.tabular import TabularPredictor, version

MLS_MODEL_DIR = Path.home().joinpath("models")


class MLSTrainableCustom:
    """
    AutoGluon을 통한 학습을 지원하는 클래스입니다.
    이 클래스는 단독으로 상속될 수 없으며 MLSModel과 함께 상속되어야 정상 동작합니다.
    `sktmls.models.contrib.EurekaModel`에서 실사용 가능합니다.
    """

    def fit(
        self,
        train_data: pd.DataFrame,
        test_data: pd.DataFrame,
        label: str,
        non_training_features: List[str] = [],
        eval_metric: str = "roc_auc",
        k: int = 30,
        time_limit: int = None,
        ensemble: bool = False,
        num_trials: int = 5,
        search_strategy: str = "auto",
        excluded_model_types: List[str] = [
            "FASTTEXT",
            "AG_TEXT_NN",
            "TRANSF",
            "custom",
        ],
    ) -> None:
        """
        AutoGluon을 통해 모델을 학습합니다.
        모델 학습이 완료되면 `self.models[0]`에 자동으로 할당되며, 이후 `predict` 함수에서 참조될 수 있습니다.
        ## 참고 사항
        - AutoGluon이 자동으로 분류 문제인지 회귀 문제인지 판단합니다.
        - 성능 지표를 설정할 수 있으나, 분류 문제의 경우 기본값인 `roc_auc` 사용을 권장합니다.
        - 회귀 문제에 분류 성능 지표를 세팅하거나 분류 문제에 회귀 성능 지표를 세팅하면 에러가 발생합니다.
        ## Args
        - train_data: (`pandas.DataFrame`) 학습에 사용할 데이터 프레임
        - test_data: (`pandas.DataFrame`) 모델 성능 측정을 위한 테스트 데이터 프레임
        - label: (str) train_data 내 라벨 컬럼 이름
        - non_training_features: (optional) (str) 학습에서 제외할 피쳐 이름 리스트. 후처리 전용 피쳐 등을 명세할 때 사용 가능 (기본값: [])
        - eval_metric: (optional) (str) 성능 지표 (기본값: `roc_auc`)
            - 분류 모델의 경우 가능한 값: `accuracy`|`balanced_accuracy`|`f1`|`f1_macro`|`f1_micro`|`f1_weighted`|`roc_auc`|`average_precision`|`precision`|`precision_macro`|`precision_micro`|`precision_weighted`|`recall`|`recall_macro`|`recall_micro`|`recall_weighted`|`log_loss`|`pac_score`
            - 회귀 모델의 경우 가능한 값: `root_mean_squared_error`|`mean_squared_error`|`mean_absolute_error`|`median_absolute_error`|`r2`
        - k: (optional) (int) Feature importance를 바탕으로 선택할 상위 피쳐의 개수 (기본값: 20)
        - time_limit: (optional) (int) 학습 시간 제한 시간 (단위: 초). n개의 모델을 학습하는 경우 1/n초씩 사용. None인 경우 무제한 (기본값: None)
        - ensemble: (optional) (bool) 앙상블 모델 학습 여부. 추론 시간이 길어지므로 학습 후 벤치마크 필수 (기본값: False)
        - num_trials: (optional) (int) Number of hpo trials you want to perform
        - search_strategy: (optional) (str) Search algorithm used by hpo experiment. ‘auto’: Random search. ‘random’: Random search. ‘bayes’: Bayes Optimization. Only supported by Ray Tune backend.
        - excluded_model_types: (optional) (List[str]) Banned subset of model types to avoid training during fit(), even if present in hyperparameters. Reference hyperparameters documentation for what models correspond to each value.
        ## Example
        features = [...]
        preprocess_logic = {...}
        postprocess_logic = {...}
        # 학습을 직접 할 것이므로 `model`을 `None`으로 할당합니다.
        my_model_v1 = GenericLogicModel(
            model=None,
            model_name="my_model",
            model_version="v1",
            features=features,
            preprocess_logic=preprocess_logic,
            postprocess_logic=postprocess_logic,
            predict_fn="predict_proba"
        )
        # 학습 및 테스트 데이터 준비
        train_data, test_data = train_test_split(df, test_size=0.2, random_state=0)
        # 학습
        my_model_v1.fit(
            train_data=train_data,
            test_data=test_data,
            label="some_label"
        )
        # 성능 확인
        print(my_model_v1.performance)
        print(my_model_v1.feature_importance)
        # predict 테스트
        print(my_model_v1.predict(test_feature_values, pf_client=pf_client, graph_client=graph_client))
        # 배포
        model_registry = ModelRegistry(env=MLSENV.STG, runtime_env=MLSRuntimeENV.YE)
        model_registry.save(my_model_v1)
        """
        self.model_lib = "autogluon"
        self.model_lib_version = version.__version__
        self.metric = eval_metric
        self.label = label
        self.non_training_features = non_training_features
        self.preprocess_logic = {"merge": [{"var": f} for f in self.features if f not in non_training_features]}
        self.num_trials = num_trials
        self.search_strategy = search_strategy
        self.excluded_model_types = excluded_model_types
        feature_cnt_check = [i for i in self.features if i not in self.non_training_features]
        if len(feature_cnt_check) <= k:
            self.models[0] = self._fit(train_data, eval_metric, time_limit, ensemble)
            self.models[0].delete_models(models_to_keep="best", dry_run=False)
            best_model = self.models[0].get_model_best()
            self.feature_importance = self.get_feature_importance(test_data).to_dict()
        else:
            if time_limit is None:
                self.models[0] = self._fit(train_data, eval_metric, time_limit, False)
                self.models[0].delete_models(models_to_keep="best", dry_run=False)
                best_model = self.models[0].get_model_best()
                feature_importance = self.get_feature_importance(test_data)
                top_features = feature_importance[:k].index.tolist()
                if len(self.features) > len(top_features):
                    self.features = top_features + non_training_features
                    self.preprocess_logic = {"merge": [{"var": f} for f in top_features]}
                    self.models[0] = self._fit(train_data, eval_metric, time_limit, ensemble)
                    self.feature_importance = self.get_feature_importance(test_data).to_dict()
            else:
                self.models[0] = self._fit(train_data, eval_metric, int(time_limit / 3), False)
                self.models[0].delete_models(models_to_keep="best", dry_run=False)
                best_model = self.models[0].get_model_best()
                feature_importance = self.get_feature_importance(test_data)
                top_features = feature_importance[:k].index.tolist()
                if len(self.features) > len(top_features):
                    self.features = top_features + non_training_features
                    self.preprocess_logic = {"merge": [{"var": f} for f in top_features]}
                    self.models[0] = self._fit(train_data, eval_metric, time_limit - int(time_limit / 3), ensemble)
                    self.feature_importance = self.get_feature_importance(test_data).to_dict()
        self.performance = self.evaluate(test_data)
        self.model_info_origin = self.models[0].info()
        self.model_info = self.model_info_origin["model_info"][best_model]
        self.model_info.pop("path", None)
        self.model_info.pop("feature_metadata", None)
        self.model_info.pop("stacker_info", None)
        self.model_info.pop("bagged_info", None)
        self.model_info.pop("children_info", None)
        self.models[0].save_space(remove_data=True, remove_fit_stack=False, requires_save=False, reduce_children=False)
        self.models[0]._trainer.reset_paths = True

    def _fit(
        self,
        train_data: pd.DataFrame,
        eval_metric: str,
        time_limit: int,
        ensemble: bool,
    ):
        columns = [f for f in self.features if f not in self.non_training_features] + [self.label]
        predictor = TabularPredictor(
            label=self.label, eval_metric=eval_metric, path=MLS_MODEL_DIR.joinpath(self.model_name, self.model_version)
        )
        return predictor.fit(
            train_data=train_data[columns],
            presets="good_quality_faster_inference_only_refit" if ensemble else "medium_quality_faster_train",
            time_limit=time_limit,
            excluded_model_types=self.excluded_model_types,
        )

    def batch_prediction(self, batch_data: pd.DataFrame) -> pd.DataFrame:
        """
        AutoGluon을 통해 학습한 모델의 배치 예측을 지원합니다.
        """
        assert self.predict_fn in [
            "predict",
            "predict_proba",
        ], "배치 예측 시 `predict_fn`은 predict, predict_proba 중 하나의 값이어야 합니다."

        if self.predict_fn == "predict":
            return self.models[0].predict(batch_data)
        elif self.predict_fn == "predict_proba":
            return self.models[0].predict_proba(batch_data)

    def evaluate(self, test_data: pd.DataFrame) -> Dict[str, float]:
        """
        AutoGluon을 통해 학습한 모델의 성능을 계산합니다.
        ## 참고 사항
        - `fit` 함수를 통해 학습이 된 경우에만 정상적으로 동작합니다.
        - `fit` 함수에서는 모델 학습 후 한 차례 본 함수를 실행하여 `self.performance`에 저장합니다.
        ## Args
        - test_data: (optional) (`pandas.DataFrame`) 모델 성능 측정을 위한 테스트 데이터 프레임 (기본값: None)
        ## Example
        # 학습 및 테스트 데이터 준비
        train_data, test_data = train_test_split(df, test_size=0.2, random_state=0)
        # 학습
        my_model_v1.fit(
            train_data=train_data,
            test_data=test_data,
            label="some_label"
        )
        # 성능 계산
        print(my_model_v1.evaluate(test_data))
        """
        columns = [f for f in self.features if f not in self.non_training_features] + [self.label]
        return self.models[0].evaluate(test_data[columns], silent=True)

    def get_feature_importance(self, test_data: pd.DataFrame) -> pd.Series:
        """
        AutoGluon을 통해 학습한 모델의 피쳐 중요도를 계산하여 `pandas.Series` 형식으로 리턴합니다.
        ## 참고 사항
        - `fit` 함수를 통해 학습이 된 경우에만 정상적으로 동작합니다.
        - `fit` 함수에서는 모델 학습 후 한 차례 본 함수를 실행하여 `self.feature_importance`에 저장합니다.
        ## Args
        - test_data: (optional) (`pandas.DataFrame`) 모델 성능 측정을 위한 테스트 데이터 프레임 (기본값: None)
        ## Example
        # 학습 및 테스트 데이터 준비
        train_data, test_data = train_test_split(df, test_size=0.2, random_state=0)
        # 학습
        my_model_v1.fit(
            train_data=train_data,
            test_data=test_data,
            label="some_label"
        )
        # 성능 계산
        print(my_model_v1.get_feature_importance(test_data))
        """
        columns = [f for f in self.features if f not in self.non_training_features] + [self.label]
        return self.models[0].feature_importance(test_data[columns], silent=True)["importance"]

    def set_mms_path(self) -> None:
        """
        MMS에서의 정상적인 inference를 위한 path 업데이트 함수로 내부 호출 용도입니다.
        """
        trainer = self.models[0]._trainer
        for model_name in trainer.get_model_names():
            trainer.set_model_attribute(
                model_name, "path", f"/models/{self.model_name}/{self.model_version}/models/{model_name}/"
            )

    def set_local_path(self) -> None:
        """
        로컬 환경에서의 정상적인 inference를 위한 path 업데이트 함수로 내부 호출 용도입니다.
        """
        trainer = self.models[0]._trainer
        for model_name in trainer.get_model_names():
            trainer.set_model_attribute(
                model_name, "path", f"{MLS_MODEL_DIR}/{self.model_name}/{self.model_version}/models/{model_name}/"
            )

    def persist_models(self) -> None:
        """
        모델 캐시를 위한 함수로 내부 호출 용도입니다.
        """
        self.models[0].persist_models()

    def unpersist_models(self) -> None:
        """
        모델 캐시 만료를 위한 함수로 내부 호출 용도입니다.
        """
        self.models[0].unpersist_models()

    def get_model_names_persisted(self) -> List[str]:
        """
        캐시된 모델 이름 조회를 위한 함수로 내부 호출 용도입니다.
        """
        return self.models[0].get_model_names_persisted()



import numpy as np
import lightgbm
import catboost

from sktmls import MLSRuntimeENV
from sktmls.apis import MLSProfileAPIClient, MLSGraphAPIClient
from sktmls.dynamodb import DynamoDBClient
from sktmls.models.contrib import ConditionalGenericLogicModel


class TestConditionalGenericLogicModel:
    def test_000_conditional_generic_logic_model0(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.6]))
        mocker.patch("catboost.CatBoostClassifier.predict_proba", return_value=np.array([0.2]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        cglm = ConditionalGenericLogicModel(
            models=[
                lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                catboost.CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            ],
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logics=[
                {
                    "merge": [
                        {"var": ["feature1", 0.0]},
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                        {
                            "first": [
                                {"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]},
                                1,
                            ]
                        },
                    ]
                },
                {
                    "merge": [
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [0, 2]]},
                    ]
                },
            ],
            postprocess_logics=[
                {
                    "if": [
                        {">": [{"var": "y"}, 0.5]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
                {
                    "if": [
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.1]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
            ],
            predict_fns=["predict", "predict_proba"],
            model_selection_logic={"abs": 0},
        )

        assert (
            cglm._select_model(
                [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
                [1],
                MLSProfileAPIClient(client_id="a", apikey="a"),
                MLSGraphAPIClient(),
                DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
            )
            == 0
        )

        assert cglm._preprocess(
            0,
            [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [2.0, 1.0, 4.0, 2.0, 4.0, 300.0]

        assert cglm.predict(
            x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            keys=[1],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == {
            "items": [
                {
                    "id": "PROD001",
                    "name": "제품1",
                    "type": "model0_type",
                    "props": {"context_id": "context_feature1", "test_score": 0.6},
                }
            ]
        }

    def test_001_conditional_generic_logic_model1(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.6]))
        mocker.patch("catboost.CatBoostClassifier.predict_proba", return_value=np.array([0.2]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        cglm = ConditionalGenericLogicModel(
            models=[
                lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                catboost.CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            ],
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logics=[
                {
                    "merge": [
                        {"var": ["feature1", 0.0]},
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                        {
                            "first": [
                                {"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]},
                                1,
                            ]
                        },
                    ]
                },
                {
                    "merge": [
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [0, 2]]},
                    ]
                },
            ],
            postprocess_logics=[
                {
                    "if": [
                        {">": [{"var": "y"}, 0.5]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
                {
                    "if": [
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.1]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
            ],
            predict_fns=["predict", "predict_proba"],
            model_selection_logic={"abs": 1},
        )

        assert (
            cglm._select_model(
                [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
                [1],
                MLSProfileAPIClient(client_id="a", apikey="a"),
                MLSGraphAPIClient(),
                DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
            )
            == 1
        )

        assert cglm._preprocess(
            1,
            [2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            [1],
            MLSProfileAPIClient(client_id="a", apikey="a"),
            MLSGraphAPIClient(),
            DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == [1.0, 4.0, 1.0, 3.0]

        assert cglm.predict(
            x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            keys=[1],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == {
            "items": [
                {
                    "id": "PROD002",
                    "name": "제품2",
                    "type": "model1_type",
                    "props": {"context_id": "context_feature2", "test_score": 0.2},
                }
            ]
        }

    def test_002_conditional_generic_logic_model_none(self, mocker):
        mocker.patch("lightgbm.basic.Booster.predict", return_value=np.array([0.6]))
        mocker.patch("catboost.CatBoostClassifier.predict_proba", return_value=np.array([0.2]))
        mocker.patch("sktmls.apis.MLSProfileAPIClient.get_item_profile", return_value={"test_dimension": 300})
        mocker.patch(
            "sktmls.apis.MLSGraphAPIClient.list_vertices",
            return_value=[{"id": "test001", "label": "testlabel", "properties": {"hello": "world"}}],
        )
        mocker.patch("sktmls.dynamodb.DynamoDBClient.get_item", return_value={"id": "ITEM0001", "age": 3})

        cglm = ConditionalGenericLogicModel(
            models=[
                lightgbm.basic.Booster(train_set=lightgbm.Dataset(np.array([[1, 2]]), np.array([1]))),
                catboost.CatBoostClassifier(custom_loss=["Accuracy"], random_seed=42, logging_level="Silent"),
            ],
            model_name="test_model",
            model_version="test_version",
            features=["feature1", "feature2", "feature3", "emb_feature1", "context_feature1", "context_feature2"],
            preprocess_logics=[
                {
                    "merge": [
                        {"var": ["feature1", 0.0]},
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [1, 3]]},
                        {
                            "first": [
                                {"pf": ["item", "test_profile", {"var": "additional_keys.0"}, ["test_dimension"]]},
                                1,
                            ]
                        },
                    ]
                },
                {
                    "merge": [
                        {"if": [{"==": [{"var": ["feature2", "N"]}, "Y"]}, 1, 0]},
                        {"float": [{"var": ["feature3", "0.0"]}]},
                        {"pick": [{"var": ["emb_feature1", [0.0] * 4]}, [0, 2]]},
                    ]
                },
            ],
            postprocess_logics=[
                {
                    "if": [
                        {">": [{"var": "y"}, 0.5]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model0_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
                {
                    "if": [
                        {">": [{"var": "y"}, 0.3]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD001",
                                        "name",
                                        "제품1",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {">": [{"var": "y"}, 0.1]},
                        {
                            "list": [
                                {
                                    "dict": [
                                        "id",
                                        "PROD002",
                                        "name",
                                        "제품2",
                                        "type",
                                        "model1_type",
                                        "props",
                                        {
                                            "dict": [
                                                "test_score",
                                                {"var": "y"},
                                                "context_id",
                                                {
                                                    "if": [
                                                        {"==": [{"var": ["context_feature2", "N"]}, "Y"]},
                                                        "context_feature2",
                                                        {"==": [{"var": ["context_feature1", "N"]}, "Y"]},
                                                        "context_feature1",
                                                        "context_default",
                                                    ]
                                                },
                                            ]
                                        },
                                    ]
                                }
                            ]
                        },
                        {"list": []},
                    ]
                },
            ],
            predict_fns=["predict", "predict_proba"],
            model_selection_logic={"list": []},
        )

        assert cglm.predict(
            x=[2.0, "Y", "4.0", [1, 2, 3, 4], "Y", "Y"],
            keys=[1],
            pf_client=MLSProfileAPIClient(client_id="a", apikey="a"),
            graph_client=MLSGraphAPIClient(),
            dynamodb_client=DynamoDBClient(runtime_env=MLSRuntimeENV.MMS),
        ) == {"items": []}
